import numpy
import torch

try:
    import cupy
except ModuleNotFoundError:
    pass

__all__ = [
    "get_backend",
    "get_vector_type",
]


def get_backend(vector):
    
    if vector.whoami == "VectorTorch":
        backend = torch
    elif vector.whoami == "VectorNumpy":
        backend = numpy
    elif vector.whoami == "VectorCupy":
        backend = cupy
        
    return backend


def get_vector_type(vector):
    if vector.whoami == "VectorTorch":
        return torch.Tensor
    elif vector.whoami == "VectorNumpy":
        return numpy.ndarray
    elif vector.whoami == "VectorCupy":
        return cupy.ndarray



import numpy as np

from occamypy import Operator
from ..vector import VectorNumpy


class FFT(Operator):
    """
    N-dimensional Fast Fourier Transform for complex input.
    
    :param model : :obj:`Vector`
        Domain vector
    :param axes : :obj:`tuple`, optional
        Dimension along which FFT is computed (all by default)
    :param nfft : :obj:`tuple`, optional
        Number of samples in Fourier Transform for each direction (same as model by default)
    :param sampling : :obj:`tuple`, optional
        Sampling steps on each axis (1. by default)
    """
    
    def __init__(self, model, axes=None, nfft=None, sampling=None):
        
        if axes is None:
            axes = tuple(range(model.ndim))
        elif not isinstance(axes, tuple) and model.ndim == 1:
            axes = (axes,)
        if nfft is None:
            nfft = model.shape
        elif not isinstance(nfft, tuple) and model.ndim == 1:
            nfft = (nfft,)
        if sampling is None:
            sampling = tuple([1.] * model.ndim)
        elif not isinstance(sampling, tuple) and model.ndim == 1:
            sampling = (sampling,)
        
        if len(axes) != len(nfft) != len(sampling):
            raise ValueError('axes, nffts, and sampling must have same number of elements')

        self.axes = axes
        self.nfft = nfft
        self.sampling = sampling
        
        self.fs = [np.fft.fftfreq(n, d=s) for n, s in zip(nfft, sampling)]
        
        dims_fft = np.asarray(model.shape)
        for a, n in zip(self.axes, self.nfft):
            dims_fft[a] = n
        
        super(FFT, self).__init__(domain=VectorNumpy(np.zeros(model.shape, dtype=np.complex)),
                                  range=VectorNumpy(np.zeros(shape=dims_fft, dtype=np.complex)))
    
    def __str__(self):
        return 'numpyFFT'
    
    def forward(self, add, model, data):
        self.checkDomainRange(model, data)
        if not add:
            data.zero()
        modelNd = model.getNdArray()
        dataNd = data.getNdArray()
        dataNd += np.fft.fftn(modelNd, s=self.nfft, axes=self.axes, norm='ortho')
        return
    
    def adjoint(self, add, model, data):
        self.checkDomainRange(model, data)
        if not add:
            model.zero()
        modelNd = model.getNdArray()
        dataNd = data.getNdArray()
        # here we need to separate the computation and use np.take for handling nfft > model.shape
        temp = np.fft.ifftn(dataNd, s=self.nfft, axes=self.axes, norm='ortho')
        for a in self.axes:
            temp = np.take(temp, range(self.domain.shape[a]), axis=a)
        modelNd += temp
        return



import os
from setuptools import setup, find_packages

NAME = "occamypy"
DESCRIPTION = "OccamyPy. An object-oriented optimization library for small- and large-scale problems."
URL = "https://github.com/fpicetti/occamypy"
EMAIL = "francesco.picetti@polimi.it"
AUTHOR = "Ettore Biondi, Guillame Barnier, Robert Clapp, Francesco Picetti, Stuart Farris"
REQUIRES_PYTHON = ">=3.6.0"
PROJECT_ROOT = os.path.abspath(os.path.dirname(__file__))


def load_readme():
    readme_path = os.path.join(PROJECT_ROOT, "README.md")
    with open(readme_path, encoding="utf-8") as f:
        return f"\n{f.read()}"


def load_version():
    context = {}
    with open(os.path.join(PROJECT_ROOT, "occamypy", "__version__.py")) as f:
        exec(f.read(), context)
    return context["__version__"]


setup(name='occamypy',
      version=load_version(),
      url="https://github.com/fpicetti/occamypy",
      description='An Object-Oriented Optimization Framework for Large-Scale Inverse Problems',
      long_description=load_readme(),
      long_description_content_type='text/markdown',
      keywords=['algebra', 'inverse problems', 'large-scale optimization'],
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'Intended Audience :: Science/Research',
          'Natural Language :: English',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Topic :: Scientific/Engineering :: Mathematics',
          'Operating System :: Unix'
      ],
      author=AUTHOR,
      author_email=EMAIL,
      install_requires=['numpy',
                        'scipy',
                        'h5py',
                        'numba',
                        'torch>=1.7.0',
                        'dask',
                        'dask-jobqueue',
                        'dask-kubernetes',
                        'matplotlib',
                        'gputil',
                        ],
      packages=find_packages(),
      zip_safe=True)



import torch
from occamypy import Operator, VectorTorch


class Module(Operator):
    """
    Encapsulate a torch.nn.Module to a operator
    """
    def __init__(self, domain, range, module, device=None):
        if not isinstance(domain, VectorTorch):
            domain = VectorTorch(domain, device=device)
        if not isinstance(range, VectorTorch):
            range = VectorTorch(range, device=device)
        
        super(Module, self).__init__(domain, range)
        
        assert isinstance(module, torch.nn.Module), "module has to be of torch.nn.Module class"
        self.module = module.to(device)
        
    def forward(self, add, model, data):
        # todo allow different devices?
        self.checkDomainRange(model, data)
        if not add:
            data.zero()
        
        # compute module output
        data[:] += self.module(model.getNdArray())
        return
    
    def parameters(self):
        return self.module.parameters()



from math import isnan

from occamypy import Solver, Problem, Operator
from occamypy.utils import ZERO


class CNN(Operator):
    def __init__(self, domain, range, architecture, name: str = None):
        super(CNN, self).__init__(domain, range)
        self.architecture = architecture
        self.name = name
    
    def forward(self, add, model, data):
        self.checkDomainRange(model, data)
        if not add:
            data.zero()
        data[:] += self.architecture(model.getNdArray())
        
    def __str__(self):
        if self.name is not None:
            return str(self.name[:8])
        else:
            return "  CNN  "
        
    @property
    def parameters(self):
        return self.architecture.parameters()
    

class DeepPriorProblem(Problem):
    def __init__(self, problem: Problem, cnn: CNN):
        """
        Cast a problem to Deep Prior paradigm

        .. math ::
            1 / 2 |A m - d|_2^2
        
        will become:
        
        .. math ::
            1 / 2 |A f(z) - d|_2^2
    
        having
        
        .. math ::
            m = f(z)
            
        where f() is the CNN, and z is a input noise tensor.
        """
        super(DeepPriorProblem, self).__init__()
    
        # necessary attributes
        self.problem = problem
        self.cnn = cnn
        self.op = problem.op
        self.model = self.cnn.domain
        self.output = self.cnn.range
        self.data = self.problem.data
        self.res = self.data.clone().zero()
        
        # other stuff
        self.dres = None
        self.grad = self.problem.grad
        self.setDefaults()
        self.linear = False
        return
    
    def objf(self, residual):
        """Method to return objective function value 1/2 | A f(z) - d |_2"""
        obj = self.problem.obj(residual)
        return obj
    
    def resf(self, model):
        """Method to return residual vector r = A f(z) - d"""
        # computing m = f(z)
        self.cnn.forward(False, model, self.output)
        # Computing A m
        if self.output.norm() != 0.:
            self.op.forward(False, self.output, self.res)
        else:
            self.res.zero()
        # Computing A m - d
        self.res.scaleAdd(self.data, 1., -1.)
        return self.res

    def gradf(self, model, res):
        """Method to return gradient vector g = A'r = A'(A f(z) - d)"""
        self.op.adjoint(False, self.grad, res)
        # TODO add autograd on cnn?
    
    
class DeepPriorSolver(Solver):
    """Deep Prior solver for Re problems"""
    
    # Default class methods/functions
    def __init__(self, optimizer, stopper, logger=None, mod_tol: float = 1e-10):
        """
        Constructor for Deep Prior Solver:
        :param optimizer: torch.opt optimizer
        :param stopper: Stopper, object to terminate inversion
        :param logger: Logger, object to write inversion log file [None]
        :param mod_tol: stop criterion for relative change of model norm
        """
        # Calling parent construction
        super(DeepPriorSolver, self).__init__()
        # Defining stopper object
        self.stopper = stopper
        # define a torch optimizer
        self.optimizer = optimizer
        # Logger object to write on log file
        self.logger = logger
        # Overwriting logger of the Stopper object
        self.stopper.logger = self.logger
        # print formatting
        self.iter_msg = "iter = %s, obj = %.5e, resnorm = %.2e"
        
        self.mod_tol = mod_tol
        
    def __del__(self):
        """Default destructor"""
        return
    
    def run(self, problem: DeepPriorProblem, verbose=False, restart=False):
        """Running Deep Prior solver"""
        self.create_msg = verbose or self.logger

        # overriding save_grad variable
        self.save_grad = False

        # Resetting stopper before running the inversion
        self.stopper.reset()
        
        # initialize all the vectors
        dp_mdl = problem.model.clone()
        dp_out = problem.output.clone()
        dp_out_old = dp_out.clone()

        if restart:
            self.restart.read_restart()
            iiter = self.restart.retrieve_parameter("iter")
            initial_obj_value = self.restart.retrieve_parameter("obj_initial")
            dp_mdl = self.restart.retrieve_vector("dp_mdl")
            dp_out = self.restart.retrieve_vector("dp_out")
            if self.create_msg:
                msg = "Restarting previous solver run from: %s" % self.restart.restart_folder
                if verbose:
                    print(msg)
                if self.logger:
                    self.logger.addToLog(msg)
        else:
            iiter = 0
            if self.create_msg:
                msg = 90 * '#' + '\n'
                msg += "\t\t\tDEEP PRIOR INVERSION log file\n\n"
                msg += "\tRestart folder: %s\n" % self.restart.restart_folder
                msg += "\tModeling Operator:\t%s\n" % problem.op
                msg += "\tDeep Prior CNN   :\t%s\n" % problem.cnn
                msg += 90 * '#' + '\n'
                if verbose:
                    print(msg.replace(" log file", ""))
                if self.logger:
                    self.logger.addToLog(msg)
        
        # main iteration loop
        while True:
            obj0 = problem.get_obj(dp_mdl)
            dp_out_old.copy(dp_out)
            
            if iiter == 0:
                initial_obj_value = obj0
                self.restart.save_parameter("obj_initial", initial_obj_value)
                if self.create_msg:
                    msg = self.iter_msg % (str(iiter).zfill(self.stopper.zfill),
                                           obj0,
                                           problem.get_rnorm(dp_mdl))
                    if verbose:
                        print(msg)
                    if self.logger:
                        self.logger.addToLog("\n" + msg)
        
            if isnan(obj0):
                raise ValueError("Objective function values NaN!")
            
            if obj0 <= ZERO:
                print("Objective function is numerically zero! Stop the inversion")
                break

            self.save_results(iiter, problem, force_save=False)
            
            # check objective function
            obj1 = problem.get_obj(dp_mdl)
            # optimize CNN
            obj1.backward()
            self.optimizer.step(obj1)
            
            # check output update tolerance criterion
            dp_out_norm = dp_out.norm()
            chng_norm = dp_out_old.scaleAdd(dp_out, 1., -1.).norm()
            if chng_norm <= self.mod_tol * dp_out_norm:
                if self.create_msg:
                    msg = "Relative output change (%.4e) norm smaller than given tolerance (%.4e)" \
                          % (chng_norm, self.mod_tol * dp_out_norm)
                    if verbose:
                        print(msg)
                    if self.logger:
                        self.logger.addToLog(msg)
                break
            
            iiter += 1
            if self.create_msg:
                msg = self.iter_msg % (str(iiter).zfill(self.stopper.zfill),
                                       obj0,
                                       problem.get_rnorm(dp_mdl))
                if verbose:
                    print(msg)
                if self.logger:
                    self.logger.addToLog("\n" + msg)

            # saving in case of restart
            self.restart.save_parameter("iter", iiter)
            self.restart.save_vector("dp_out", dp_out)

            if self.stopper.run(problem, iiter, initial_obj_value, verbose):
                break

        # writing last inverted model
        self.save_results(iiter, problem, force_save=True, force_write=True)

        # ending message and log file
        if self.create_msg:
            msg = 90 * '#' + '\n'
            msg += "\t\t\tDEEP PRIOR INVERSION log file end\n"
            msg += 90 * '#'
            if verbose:
                print(msg.replace(" log file", ""))
            if self.logger:
                self.logger.addToLog("\n" + msg)

        # Clear restart object
        self.restart.clear_restart()


if __name__ == "__main__":
    import occamypy
    import numpy as np
    import segmentation_models_pytorch as smp
    import torch
    
    # reference image
    x = np.load("./tutorials/data/monarch.npy").astype(float)
    x = occamypy.VectorTorch(x[None, None]).scale(1 / 255)
    
    # noisy image
    y = x.clone() + x.clone().randn(40.)
    
    # input tensor
    z = occamypy.VectorTorch((1, 64, *x.shape[2:])).randn()
    
    # deep prior cnn
    f = CNN(domain=z, range=x, name="unet_R18",
            architecture=smp.Unet(in_channels=z.shape[1],
                                  classes=x.shape[1],
                                  encoder_name="resnet18",
                                  encoder_weights=None,
                                  activation=None))
    
    x_hat = f*z

    # linear problem: 1/2 |Ax - y|_2^2
    lin_problem = occamypy.LeastSquares(x.clone(), y, occamypy.Identity(x))
    # deep prior formulation: x = f(z)
    dip_problem = DeepPriorProblem(lin_problem, cnn=f)
    
    DIP = DeepPriorSolver(
        optimizer=torch.optim.Adam(f.parameters, lr=0.001),
        stopper=occamypy.BasicStopper(1000),
        logger=None,
    )
    
    DIP.run(dip_problem, verbose=True)
